#include <QtGui>
#include <iostream>
#include <cstdlib>
#include <ctime>
#include "bsdetectionwidget.h"

using namespace std;



const int BSDetectionWidget::MAX_WIDTH_TUNING = 400;
const int BSDetectionWidget::DETECTION_LACKS_TUNING = 20;
const int BSDetectionWidget::DEFAULT_PEN_WIDTH = 1;



BSDetectionWidget::BSDetectionWidget (QWidget *parent)
{
  Q_UNUSED (parent);

  // Sets default user interface parameters
  setFocus ();
  grabKeyboard ();
  fixed = false;

  // Sets initial values for the gradient map
  gMap = NULL;

  // Initializes the auxiliary views
  accuview = NULL;
  strucview = NULL;
  profileview = NULL;

  alternate = 0;
  verbose = false;
  capture = false;
  exam = -1;
}


BSDetectionWidget::~BSDetectionWidget ()
{
  if (accuview != NULL) delete accuview;
  if (strucview != NULL) delete strucview;
}


QSize BSDetectionWidget::openImage (const QString &fileName, int type)
{
  QSize newSize (0, 0);
  loadedImage.load (fileName);
  width = loadedImage.width ();
  height = loadedImage.height ();
  newSize = loadedImage.size ();
 
  augmentedImage = loadedImage;
  if (gMap != NULL) delete gMap;
  gMap = new VMap (width, height, getBitmap (augmentedImage), type);
  detector.setGradientMap (gMap);
 
  update ();
  if (profileview != NULL) profileview->setImage (&loadedImage, gMap);
  if (strucview != NULL) strucview->setGradientImage (gMap);

  return newSize;
}


int **BSDetectionWidget::getBitmap (const QImage &image)
{
  int w = image.width ();
  int h = image.height ();
  
  int **tabImage = new int*[h];
  for (int i = 0; i < h; i++)
  {
    tabImage[i] = new int[w];
    for(int j = 0; j < w; j++)
    {
      QColor c = QColor (image.pixel (j, h - i - 1));
      tabImage[i][j] = c.value ();
    }
  }
  return tabImage;
}


bool BSDetectionWidget::saveAugmentedImage (const QString &fileName,
                                            const char *fileFormat)
{
  QImage aImage = augmentedImage;
  return (aImage.save (fileName, fileFormat));
}


void BSDetectionWidget::clearImage ()
{
  augmentedImage.fill (qRgb (255, 255, 255));
  update ();
}


void BSDetectionWidget::paintEvent (QPaintEvent *)
{
  QPainter painter (this);
  painter.drawImage (QPoint (0, 0), augmentedImage);
}


void BSDetectionWidget::storeExtractedSegment ()
{
  BlurredSegment *bs = detector.getBlurredSegment (1);
  if (bs != NULL)
  {
    ExtractedSegment es;
    es.bs = bs;
    es.p1 = p1;
    es.p2 = p2;
    extractedSegments.push_back (es);
    detector.preserveFormerBlurredSegment ();
  }
}


void BSDetectionWidget::displayExtractedSegments ()
{
  augmentedImage = loadedImage;
  QPainter painter (&augmentedImage);
  if (! extractedSegments.empty ())
  {
    vector<ExtractedSegment>::iterator it = extractedSegments.begin ();
    while (it != extractedSegments.end ())
    {
      vector<Pt2i> bnd;
      DigitalStraightSegment *dss = it->bs->getSegment ();
      dss->getBoundPoints (bnd, false, 0, 0, width, height);
      drawListOfPixels (bnd, Qt::green, painter);
      dss->getBoundPoints (bnd, true, 0, 0, width, height);
      drawListOfPixels (bnd, Qt::green, painter);
      bnd = it->bs->getAllPoints ();
      drawListOfPixels (bnd, Qt::blue, painter);
      drawLine (it->p1, it->p2, Qt::white, painter);
      it ++;
    }
  }
  update (QRect (QPoint (0, 0), QPoint (width, height)));
}


void BSDetectionWidget::clearExtractedSegments ()
{
  vector<ExtractedSegment>::iterator it = extractedSegments.begin ();
  while (it != extractedSegments.end ())
    delete ((it++)->bs->getSegment ());
  extractedSegments.clear ();
}


void BSDetectionWidget::closeAccuAnalyzer ()
{
  if (accuview != NULL)
  {
    accuview->close ();
    delete accuview;
    accuview = NULL;
  }
}


void BSDetectionWidget::closePixelAnalyzer ()
{
  if (strucview != NULL)
  {
    strucview->close ();
    delete strucview;
    strucview = NULL;
  }
}


void BSDetectionWidget::closeProfileAnalyzer ()
{
  if (profileview != NULL)
  {
    profileview->close ();
    delete profileview;
    profileview = NULL;
  }
}


void BSDetectionWidget::switchAccuAnalyzer ()
{
  if (accuview != NULL)
  {
    accuview->close ();
    delete accuview;
    accuview = NULL;
  }
  else
  {
    accuview = new BSAccumulatorView (&detector);
    accuview->show ();
  }
}


void BSDetectionWidget::switchPixelAnalyzer ()
{
  if (strucview != NULL)
  {
    strucview->close ();
    delete strucview;
    strucview = NULL;
  }
  else
  {
    strucview = new BSStructureView (&loadedImage, &detector);
    strucview->setGradientImage (gMap);
    strucview->show ();
  }
}


void BSDetectionWidget::switchProfileAnalyzer ()
{
  if (profileview != NULL)
  {
    profileview->close ();
    delete profileview;
    profileview = NULL;
  }
  else
  {
    profileview = new BSProfileView ();
    profileview->setImage (&loadedImage, gMap);
    if (! p1.equals (p2)) profileview->buildScans (p1, p2);
    profileview->show ();
  }
}


void BSDetectionWidget::mousePressEvent (QMouseEvent *event)
{
  this->p1 = Pt2i (event->pos().x (), height - 1 - event->pos().y());
}


void BSDetectionWidget::mouseReleaseEvent (QMouseEvent *event)
{
  fixed = true;
  this->p2 = Pt2i (event->pos().x (), height - 1 - event->pos().y());
alternate = 0;
  cerr << "p1 defined: " << p1.x () << " " << p1.y () << endl;
  cerr << "p2 defined: " << p2.x () << " " << p2.y () << endl;
  extract ();
}


void BSDetectionWidget::mouseMoveEvent (QMouseEvent *event)
{
  fixed = false;
  this->p2 = Pt2i (event->pos().x (), height - 1 - event->pos().y ());
  if (verbose) cerr << "(" << p1.x () << ", " << p1.y () << ") ("
                    << p2.x () << ", " << p2.y () << ")" << endl;
  if (p1.manhattan (p2) > 5
      && (width > p2.x() && height > p2.y()
          && p2.x() > 0 && p2.y() > 0))
  {
    extract ();
  }
}


void BSDetectionWidget::keyPressEvent (QKeyEvent *event)
{
  if (isActiveWindow ()) switch (event->key ())
  {
    case Qt::Key_U :
      cerr << "p1 update: " << p1.x () << " " << p1.y () << endl;
      cerr << "p2 update: " << p2.x () << " " << p2.y () << endl;
      extract (true);
      break;

    case Qt::Key_M : // Multi-selection switch
      if (event->modifiers () & Qt::ShiftModifier)
      {
        detector.switchMultiSelection ();
        cout << "Multi-selection "
             << (detector.isMultiSelection () ? "on" : "off") << endl;
      }
      else extractAll ();
      break;

    case Qt::Key_L : // Next multi-selected segment
      displayNext (event->modifiers () & Qt::ShiftModifier ? -1 : 1);
      cout << "Selection du segment " << exam << endl;
      break;

    case Qt::Key_Y : // Initial detection extension limitation
      detector.switchInitialBounding ();
      extract (true);
      cout << "Initial step max extension = "
           << detector.initialDetectionMaxExtent () << endl;
      break;

    case Qt::Key_D : // Density test at initial step
      detector.switchDensityTest ();
      extract (true);
      cout << "Density test : "
           << (detector.isSetDensityTest () ? "on" : "off") << endl;
      break;

    case Qt::Key_W : // Input max width
      detector.setInputMaxWidth (detector.getInputMaxWidth () +
        (event->modifiers () & Qt::ShiftModifier ? -1 : 1));
      extract (true);
      cout << "Input max width = " << detector.getInputMaxWidth () << endl;
      break;

    case Qt::Key_K : // Output blurred segment min size
      detector.setBSminSize (detector.getBSminSize () +
        (event->modifiers () & Qt::ShiftModifier ? -1 : 1));
      extract (true);
      cout << "Output BS min size = " << detector.getBSminSize () << endl;
      break;

    case Qt::Key_H : // Pixel lack tolerence
      detector.setPixelLackTolerence (detector.getPixelLackTolerence () +
        (event->modifiers () & Qt::ShiftModifier ? -1 : 1));
      extract (true);
      cout << "Lack tolerence = " << detector.getPixelLackTolerence ()
           << " pixels" << endl;
      break;

    case Qt::Key_N : // Automatic detection grid resolution
      detector.setAutoGridResolution (detector.getAutoGridResolution () +
        (event->modifiers () & Qt::ShiftModifier ? -1 : 1));
      cout << "Auto grid resolution = "
           << detector.getAutoGridResolution () << " pixels" << endl;
      break;

    case Qt::Key_Q :
      detector.switchAutoRestart ();
      extract (true);
      cout << "Segment continuation after = "
           << detector.getRestartOnLack () << " pixels" << endl;
      break;

    case Qt::Key_T :
      detector.toggleThinning ();
      extract (true);
      cout << "Thinning "
           << (detector.isThinningActivated () ? "on" : "off") << endl;
      break;

    case Qt::Key_X :
      detector.switchAutoWidth ();
      cout << "Final step max width " << (detector.autoWidthOn () ?
              "fitted to initial segment" : "left unchanged") << endl;
      extract (true);
      break;

    case Qt::Key_S :
      detector.switchDynamicScans ();
      cout << (detector.dynamicScansOn () ?
               "Dynamic scans" : "Static scans") << endl;
      extract (true);
      break;

    case Qt::Key_O :
      detector.switchOrthoScans ();
      cout << (detector.orthoScansOn () ?
               "Orthographic scans" : "Directional scans") << endl;
      extract (true);
      break;

    case Qt::Key_F :
      detector.switchFiltering (1);
      cout << "Final filtering "
           << (detector.isFiltering (1) ? "on" : "off") << endl;
      extract (true);
      break;

    case Qt::Key_P :
      if (event->modifiers () & Qt::ShiftModifier)
      {
        capture = true;
        cout << "Next extraction will be shot in capture.png" << endl;
      }
      else
      {
        detector.switchFiltering (0);
        cout << "Pre-filtering "
             << (detector.isFiltering (0) ? "on" : "off") << endl;
        extract (true);
      }
      break;

    case Qt::Key_G :
      detector.switchConnectivityConstraint ();
      cout << "Connectivity constraint "
           << (detector.isSetConnectivityConstraint () ? "on" : "off") << endl;
      extract (true);
      break;

    case Qt::Key_C :
      if (event->modifiers () & Qt::ControlModifier)
        storeExtractedSegment ();
      break;

    case Qt::Key_V :
      if (event->modifiers () & Qt::ControlModifier)
        displayExtractedSegments ();
      else switchVerbose ();
      break;

    case Qt::Key_Z :
      if (event->modifiers () & Qt::ControlModifier)
        clearExtractedSegments ();
      break;

    case Qt::Key_E :
      if (event->modifiers () & Qt::ShiftModifier)
        detector.invertEdgeDirection ();
      else detector.switchEdgeDirectionConstraint ();
      switch (detector.edgeDirectionConstraint ())
      {
        case 0 :
          cout << "Stroke detection" << endl;
          break;
        case 1 :
          cout << "Edge detection" << endl;
          break;
        case -1 :
          cout << "Opposite edge detection" << endl;
          break;
      }
      extract (true);
      break;

    case Qt::Key_1 :
      switchPixelAnalyzer ();
      break;

    case Qt::Key_2 :
      switchAccuAnalyzer ();
      break;

    case Qt::Key_3 :
      switchProfileAnalyzer ();
      break;

    case Qt::Key_8 :
      alternateTest ();
      break;

    case Qt::Key_9 :
      performanceTest ();
      break;

    case Qt::Key_0 :
      localTest ();
      break;
  }
  else if (strucview != NULL && strucview->isActiveWindow ())
  {
    if (strucview->processKeyEvent (event)) extract (true);
  }
  else if (accuview != NULL && accuview->isActiveWindow ())
  {
    if (accuview->processKeyEvent (event)) extract (true);
  }
  else if (profileview != NULL && profileview->isActiveWindow ())
  {
    if (profileview->processKeyEvent (event)) extract (true);
  }
}


void BSDetectionWidget::drawListOfPixels (vector<Pt2i> vectPixels,
                                          QColor color, QPainter &painter)
{
  vector<Pt2i>::iterator iter = vectPixels.begin ();
  while (iter != vectPixels.end ())
  {
    Pt2i p = *iter;
    painter.setPen (QPen (color, DEFAULT_PEN_WIDTH, Qt::SolidLine,
                          Qt::RoundCap, Qt::RoundJoin));
    if (p.x() < width && p.y() < height && p.x() >= 0 && p.y() >= 0)
      painter.drawPoint (QPoint (p.x(), height - 1 - p.y()));  // dec 1
    iter ++;
   }
}


void BSDetectionWidget::drawListOfPixels (vector<Pt2i> vectPixels,
                                          QPainter &painter)
{
  vector<Pt2i>::iterator iter = vectPixels.begin ();
  while (iter != vectPixels.end ())
  {
    Pt2i p = *iter;
    painter.setPen (QPen (QBrush (loadedImage.pixel (p.x (),
                                  loadedImage.height () - 1 - p.y ())),
                          DEFAULT_PEN_WIDTH, Qt::SolidLine,
                          Qt::RoundCap, Qt::RoundJoin));
    if (p.x() < width && p.y() < height && p.x() >= 0 && p.y() >= 0)
      painter.drawPoint (QPoint (p.x(), height - 1 - p.y()));  // dec 1
    iter ++;
   }
}


void BSDetectionWidget::drawLine (const Pt2i from, const Pt2i to,
                                  QColor color, QPainter &painter)
{
  int n;
  Pt2i *pts = from.drawing (to, &n);
  painter.setPen (QPen (color, DEFAULT_PEN_WIDTH, Qt::SolidLine,
                        Qt::RoundCap, Qt::RoundJoin));
  for (int i = 0; i < n; i++)
    painter.drawPoint (QPoint (pts[i].x (),
                               height - 1 - pts[i].y ()));  // dec 1
  delete [] pts;
}


void BSDetectionWidget::extract (bool withAllDisplay)
{
  if (! p1.equals (p2))
  {
    fixed = withAllDisplay;
    extract ();
  }
}


void BSDetectionWidget::extract ()
{
  augmentedImage = loadedImage;
  QPainter painter (&augmentedImage);
  if (p1.equals (p2))
  {
    update (QRect (QPoint (0, 0), QPoint (width, height)));
    fixed = false;
    return;
  }
  drawLine (p1, p2, Qt::red, painter);

  if (detector.isMultiSelection ()) detector.multidetect (p1, p2);
  else
  {
    detector.freeMultiSelection ();
    detector.detect (p1, p2);
  }

  // Update auxiliary view if not dragging
  if (fixed)
  {
    if (profileview != NULL)
    {
      profileview->buildScans (p1, p2);
      profileview->scene()->update ();
    }
    if (accuview != NULL) accuview->scene()->update ();
    if (strucview != NULL)
    {
      strucview->setExamined (-1);
      strucview->scene()->update ();
    }
  }

  if (detector.isMultiSelection ())
  {
    vector<BlurredSegment *> mbs = detector.getBlurredSegments (1);
    vector<BlurredSegment *>::iterator it = mbs.begin ();
    while (it != mbs.end ())
    {
      DigitalStraightSegment *dss = (*it)->getSegment ();
      if (dss != NULL) 
      {
        vector<Pt2i> bnd;
        dss->getBounds (bnd, 0, 0, width, height);
        drawListOfPixels (bnd, Qt::green, painter);
      }
      it++;
    }
  }
  else
  {
    BlurredSegment *bs = detector.getBlurredSegment (1);
    if (bs != NULL)
    {
      DigitalStraightSegment *dss = bs->getSegment ();
      if (dss != NULL) 
      {
        vector<Pt2i> bnd;
        dss->getBounds (bnd, 0, 0, width, height);
        drawListOfPixels (bnd, Qt::green, painter);
      }
    }
  }

  update (QRect (QPoint (0, 0), QPoint (width, height)));
  if (verbose && fixed) displayExtractionResult ();
  fixed = false;
}


void BSDetectionWidget::extractAll ()
{
  bool formerMultiMode = detector.isMultiSelection ();
  if (! formerMultiMode) detector.switchMultiSelection ();
  augmentedImage = loadedImage;
  QPainter painter (&augmentedImage);

  detector.detectAll ();

  // Update auxiliary view
  if (profileview != NULL)
  {
    profileview->buildScans (p1, p2);
    profileview->scene()->update ();
  }
  if (accuview != NULL) accuview->scene()->update ();
  if (strucview != NULL)
  {
    strucview->setExamined (-1);
    strucview->scene()->update ();
  }

  // Update auxiliary view
  if (strucview != NULL) strucview->scene()->update ();

  vector<BlurredSegment *> mbs = detector.getBlurredSegments (1);
  cout << mbs.size () << " blurred segments detected" << endl;
  vector<BlurredSegment *>::iterator it = mbs.begin ();
  while (it != mbs.end ())
  {
/*
    DigitalStraightSegment *dss = (*it)->getSegment ();
    if (dss != NULL) 
    {
      vector<Pt2i> bnd;
      dss->getBounds (bnd, 0, 0, width, height);
      drawListOfPixels (bnd, Qt::green, painter);
    }
*/
    drawListOfPixels ((*it)->getAllPoints (), Qt::green, painter);
    it++;
  }

  update (QRect (QPoint (0, 0), QPoint (width, height)));
  if (verbose && fixed) displayExtractionResult ();
  fixed = false;
  if (! formerMultiMode) detector.switchMultiSelection ();
  exam = -1;

  if (capture)
  {
    augmentedImage.save ("capture.png");
    capture = false;
  }
}


void BSDetectionWidget::displayNext (int dir)
{
  vector<BlurredSegment *> mbs = detector.getBlurredSegments (1);
  if (mbs.empty ()) return;

  augmentedImage = loadedImage;
  QPainter painter (&augmentedImage);
  exam = (exam + dir) % ((int) (mbs.size ()));

  // Update auxiliary view
  if (strucview != NULL)
  {
    strucview->setExamined (exam);
    strucview->scene()->update ();
  }

  vector<BlurredSegment *>::iterator it = mbs.begin ();
  QColor bsCol = Qt::blue;
  int index = 0;
  while (it != mbs.end ())
  {
    if (index++ == exam) bsCol = Qt::green;
    else bsCol = Qt::blue;
/*
    DigitalStraightSegment *dss = (*it)->getSegment ();
    if (dss != NULL) 
    {
      vector<Pt2i> bnd;
      dss->getBounds (bnd, 0, 0, width, height);
      drawListOfPixels (bnd, Qt::green, painter);
    }
*/
    drawListOfPixels ((*it)->getAllPoints (), bsCol, painter);
    it++;
  }
  update (QRect (QPoint (0, 0), QPoint (width, height)));
}


void BSDetectionWidget::displayExtractionResult ()
{
  int res = detector.result ();
  if (res == BSDetector::RESULT_UNDETERMINED)
    cout << "Extraction : undetermined." << endl;
  else if (res == BSDetector::RESULT_OK)
    cout << "Extraction : OK." << endl;
  else if (res == BSDetector::RESULT_INITIAL_NO_DETECTION)
    cout << "Extraction : no initial detection (bsini == NULL)." << endl;
  else if (res == BSDetector::RESULT_INITIAL_TOO_FEW)
    cout << "Extraction : two few points at initial detection." << endl;
  else if (res == BSDetector::RESULT_INITIAL_TOO_SPARSE)
    cout << "Extraction : unsuccessful density test at initial detection."
         << endl;
  else if (res == BSDetector::RESULT_INITIAL_TOO_MANY_OUTLIERS)
    cout << "Extraction : unsuccessful filter test at initial detection."
         << endl;
  else if (res == BSDetector::RESULT_FINAL_NO_DETECTION)
    cout << "Extraction : no final detection (bsini == NULL)." << endl;
  else if (res == BSDetector::RESULT_FINAL_TOO_FEW)
    cout << "Extraction : two few points at final detection." << endl;
  else if (res == BSDetector::RESULT_FINAL_TOO_SPARSE)
    cout << "Extraction : unsuccessful density test at final detection."
         << endl;
  else if (res == BSDetector::RESULT_FINAL_TOO_MANY_OUTLIERS)
    cout << "Extraction : unsuccessful filter test at final detection."
         << endl;
}


void BSDetectionWidget::alternateTest ()
{
  if (p1.equals (p2))
  {
    cout << "Stroke undefined" << endl;
    return;
  }

  if (++alternate == 3) alternate = 0;
  if (alternate == 0)
  {
  }
  else if (alternate == 1)
  {
  }
  else if (alternate == 2)
  {
  }
//  extract (true);
} 



void BSDetectionWidget::performanceTest ()
{
/*
  if (p1.equals (p2))
  {
    cout << "Stroke undefined" << endl;
    return;
  }
  cout << "Run test" << endl;
  clock_t start = clock ();
  for (int i = 0; i < 1000; i++) detector.detect (p1, p2);
  double diff = (clock () - start) / (double) CLOCKS_PER_SEC;
  cout << "Test run : " << diff << endl;
  extract (true);
*/

  cout << "Complete extractions test" << endl;
  clock_t start = clock ();
  for (int i = 0; i < 100; i++) detector.detectAll ();
  double diff = (clock () - start) / (double) CLOCKS_PER_SEC;
  cout << "Test run : " << diff << endl;
  extract (true);
}


void BSDetectionWidget::localTest ()
{
  cout << "Run test" << endl;

/*
  p1 = Pt2i (212, 169);
  p2 = Pt2i (232, 152);
*/

  p1 = Pt2i (298, 199);
  p2 = Pt2i (279, 173);

  extract (true);
  cout << "Test run" << endl;
}
